# Standard Environment variables from AWS Lambda:
# https://docs.aws.amazon.com/lambda/latest/dg/configuration-envvars.html#configuration-envvars-runtime
AWS_LAMBDA_FUNCTION_NAME_KEY = "AWS_LAMBDA_FUNCTION_NAME"
AWS_LAMBDA_FUNCTION_VERSION_KEY = "AWS_LAMBDA_FUNCTION_VERSION"
AWS_LAMBDA_FUNCTION_MEMORY_SIZE_KEY = "AWS_LAMBDA_FUNCTION_MEMORY_SIZE"
AWS_LAMBDA_LOG_GROUP_NAME_KEY = "AWS_LAMBDA_LOG_GROUP_NAME"
AWS_LAMBDA_LOG_STREAM_NAME_KEY = "AWS_LAMBDA_LOG_STREAM_NAME"
AWS_EXECUTION_ENV_KEY = "AWS_EXECUTION_ENV"
AWS_LAMBDA_FUNCTION_HANDLER_STANDARD_KEY = "_HANDLER"

# These are custom env variable names
AWS_LAMBDA_FUNCTION_REQUEST_ID_KEY = "AWS_LAMBDA_FUNCTION_REQUEST_ID"
AWS_LAMBDA_FUNCTION_ARN_KEY = "AWS_LAMBDA_FUNCTION_ARN"
AWS_LAMBDA_FUNCTION_HANDLER_KEY = "AWS_LAMBDA_FUNCTION_HANDLER"
AWS_LAMBDA_EVENT_PAYLOAD_KEY = "AWS_LAMBDA_EVENT_PAYLOAD"
AWS_LAMBDA_EVENT_RESPONSE_LOCATION_KEY = "AWS_LAMBDA_EVENT_RESPONSE_LOCATION"


DEFAULT_AWS_LAMBDA_FUNCTION_NAME = "unknown"
